"""
This module contains all of fbs's built-in commands. They are invoked when you
run `fbs <command>` on the command line. But you are also free to import them in
your Python build script and execute them there.
"""
from fbs import path, SETTINGS, activate_profile
from fbs.builtin_commands._util import prompt_for_value, is_valid_version, \
    require_existing_project, update_json, require_frozen_app, require_installer
from fbs.cmdline import command
from fbs.resources import copy_with_filtering
from fbs.upload import _upload_repo
from fbs_runtime import FbsError
from fbs_runtime.platform import is_windows, is_mac, is_linux, is_arch_linux, \
    is_ubuntu, is_fedora
from getpass import getuser
from importlib.util import find_spec
from os import listdir, remove, unlink, mkdir
from os.path import join, isfile, isdir, islink, dirname, exists, relpath
from shutil import rmtree
from unittest import TestSuite, TextTestRunner, defaultTestLoader

import logging
import os
import subprocess
import sys

_LOG = logging.getLogger(__name__)

@command
def startproject():
    """
    Start a new project in the current directory
    """
    if exists('src'):
        raise FbsError('The src/ directory already exists. Aborting.')
    app = prompt_for_value('App name', default='MyApp')
    user = getuser().title()
    author = prompt_for_value('Author', default=user)
    has_pyqt = _has_module('PyQt5')
    has_pyside = _has_module('PySide2')
    if has_pyqt and not has_pyside:
        python_bindings = 'PyQt5'
    elif not has_pyqt and has_pyside:
        python_bindings = 'PySide2'
    else:
        python_bindings = prompt_for_value(
            'Qt bindings', choices=('PyQt5', 'PySide2'), default='PyQt5'
        )
    eg_bundle_id = 'com.%s.%s' % (
        author.lower().split()[0], ''.join(app.lower().split())
    )
    mac_bundle_identifier = prompt_for_value(
        'Mac bundle identifier (eg. %s, optional)' % eg_bundle_id,
        optional=True
    )
    mkdir('src')
    template_dir = join(dirname(__file__), 'project_template')
    template_path = lambda relpath: join(template_dir, *relpath.split('/'))
    copy_with_filtering(
        template_dir, '.', {
            'app_name': app,
            'author': author,
            'mac_bundle_identifier': mac_bundle_identifier,
            'python_bindings': python_bindings
        },
        files_to_filter=[
            template_path('src/build/settings/base.json'),
            template_path('src/build/settings/mac.json'),
            template_path('src/main/python/main.py')
        ]
    )
    print('')
    _LOG.info(
        "Created the src/ directory. If you have %s installed, you can now "
        "do:\n\n    fbs run", python_bindings
    )

@command
def run():
    """
    Run your app from source
    """
    require_existing_project()
    if not _has_module('PyQt5') and not _has_module('PySide2'):
        raise FbsError(
            "Couldn't find PyQt5 or PySide2. Maybe you need to:\n"
            "    pip install PyQt5==5.9.2 or\n"
            "    pip install PySide2==5.12.2"
        )
    env = dict(os.environ)
    pythonpath = path('src/main/python')
    old_pythonpath = env.get('PYTHONPATH', '')
    if old_pythonpath:
        pythonpath += os.pathsep + old_pythonpath
    env['PYTHONPATH'] = pythonpath
    subprocess.run([sys.executable, path(SETTINGS['main_module'])], env=env)

@command
def freeze(debug=False):
    """
    Compile your code to a standalone executable
    """
    require_existing_project()
    if not _has_module('PyInstaller'):
        raise FbsError(
            "Could not find PyInstaller. Maybe you need to:\n"
            "    pip install PyInstaller==5.13.2"
        )
    version = SETTINGS['version']
    if not is_valid_version(version):
        raise FbsError(
            'Invalid version detected in settings. It should be three\n'
            'numbers separated by dots, such as "1.2.3". You have:\n\t"%s".\n'
            'Usually, this can be fixed in src/build/settings/base.json.'
            % version
        )
    # Import respective functions late to avoid circular import
    # fbs <-> fbs.freeze.X.
    app_name = SETTINGS['app_name']
    if is_mac():
        from fbs.freeze.mac import freeze_mac
        freeze_mac(debug=debug)
        executable = 'target/%s.app/Contents/MacOS/%s' % (app_name, app_name)
    else:
        executable = join('target', app_name, app_name)
        if is_windows():
            from fbs.freeze.windows import freeze_windows
            freeze_windows(debug=debug)
            executable += '.exe'
        elif is_linux():
            if is_ubuntu():
                from fbs.freeze.ubuntu import freeze_ubuntu
                freeze_ubuntu(debug=debug)
            elif is_arch_linux():
                from fbs.freeze.arch import freeze_arch
                freeze_arch(debug=debug)
            elif is_fedora():
                from fbs.freeze.fedora import freeze_fedora
                freeze_fedora(debug=debug)
            else:
                from fbs.freeze.linux import freeze_linux
                freeze_linux(debug=debug)
        else:
            raise FbsError('Unsupported OS')
    _LOG.info(
        "Done. You can now run `%s`. If that doesn't work, see "
        "https://build-system.fman.io/troubleshooting.", executable
    )

@command
def sign():
    """
    Sign your app, so the user's OS trusts it
    """
    require_frozen_app()
    if is_windows():
        from fbs.sign.windows import sign_windows
        sign_windows()
        _LOG.info(
            'Signed all binary files in %s and its subdirectories.',
            relpath(path('${freeze_dir}'), path('.'))
        )
    elif is_mac():
        _LOG.info('fbs does not yet implement `sign` on macOS.')
    else:
        _LOG.info('This platform does not support signing frozen apps.')

@command
def installer():
    """
    Create an installer for your app
    """
    require_frozen_app()
    linux_distribution_not_supported_msg = \
        "Your Linux distribution is not supported, sorry. " \
        "You can run `fbs buildvm` followed by `fbs runvm` to start a Docker " \
        "VM of a supported distribution."
    try:
        installer_fname = SETTINGS['installer']
    except KeyError:
        if is_linux():
            raise FbsError(linux_distribution_not_supported_msg)
        raise
    out_file = join('target', installer_fname)
    msg_parts = ['Created %s.' % out_file]
    if is_windows():
        from fbs.installer.windows import create_installer_windows
        create_installer_windows()
    elif is_mac():
        from fbs.installer.mac import create_installer_mac
        create_installer_mac()
    elif is_linux():
        app_name = SETTINGS['app_name']
        if is_ubuntu():
            from fbs.installer.ubuntu import create_installer_ubuntu
            create_installer_ubuntu()
            install_cmd = 'sudo dpkg -i ' + out_file
            remove_cmd = 'sudo dpkg --purge ' + app_name
        elif is_arch_linux():
            from fbs.installer.arch import create_installer_arch
            create_installer_arch()
            install_cmd = 'sudo pacman -U ' + out_file
            remove_cmd = 'sudo pacman -R ' + app_name
        elif is_fedora():
            from fbs.installer.fedora import create_installer_fedora
            create_installer_fedora()
            install_cmd = 'sudo dnf install ' + out_file
            remove_cmd = 'sudo dnf remove ' + app_name
        else:
            raise FbsError(linux_distribution_not_supported_msg)
        msg_parts.append(
            'You can for instance install it via the following command:\n'
            '    %s\n'
            'This places it in /opt/%s. To uninstall it again, you can use:\n'
            '    %s'
            % (install_cmd, app_name, remove_cmd)
        )
    else:
        raise FbsError('Unsupported OS')
    _LOG.info(' '.join(msg_parts))

@command
def sign_installer():
    """
    Sign installer, so the user's OS trusts it
    """
    if is_mac():
        _LOG.info('fbs does not yet implement `sign_installer` on macOS.')
        return
    if is_ubuntu():
        _LOG.info('Ubuntu does not support signing installers.')
        return
    require_installer()
    if is_windows():
        from fbs.sign_installer.windows import sign_installer_windows
        sign_installer_windows()
    elif is_arch_linux():
        from fbs.sign_installer.arch import sign_installer_arch
        sign_installer_arch()
    elif is_fedora():
        from fbs.sign_installer.fedora import sign_installer_fedora
        sign_installer_fedora()
    _LOG.info('Signed %s.', join('target', SETTINGS['installer']))

@command
def repo():
    """
    Generate files for automatic updates
    """
    require_existing_project()
    if not _repo_is_supported():
        raise FbsError('This command is not supported on this platform.')
    app_name = SETTINGS['app_name']
    pkg_name = app_name.lower()
    try:
        gpg_key = SETTINGS['gpg_key']
    except KeyError:
        raise FbsError(
            'GPG key for code signing is not configured. You might want to '
            'either\n'
            '    1) run `fbs gengpgkey` or\n'
            '    2) set "gpg_key" and "gpg_pass" in src/build/settings/.'
        )
    if is_ubuntu():
        from fbs.repo.ubuntu import create_repo_ubuntu
        if not SETTINGS['description']:
            _LOG.info(
                'Hint: Your app\'s "description" is empty. Consider setting it '
                'in src/build/settings/linux.json.'
            )
        create_repo_ubuntu()
        _LOG.info(
            'Done. You can test the repository with the following commands:\n'
            '    echo "deb [arch=amd64] file://%s stable main" '
                '| sudo tee /etc/apt/sources.list.d/%s.list\n'
            '    sudo apt-key add %s\n'
            '    sudo apt-get update\n'
            '    sudo apt-get install %s\n'
            'To revert these changes:\n'
            '    sudo dpkg --purge %s\n'
            '    sudo apt-key del %s\n'
            '    sudo rm /etc/apt/sources.list.d/%s.list\n'
            '    sudo apt-get update',
            path('target/repo'), pkg_name,
            path('src/sign/linux/public-key.gpg'), pkg_name, pkg_name, gpg_key,
            pkg_name,
            extra={'wrap': False}
        )
    elif is_arch_linux():
        from fbs.repo.arch import create_repo_arch
        create_repo_arch()
        _LOG.info(
            "Done. You can test the repository with the following commands:\n"
            "    sudo cp /etc/pacman.conf /etc/pacman.conf.bu\n"
            "    echo -e '\\n[%s]\\nServer = file://%s' "
                "| sudo tee -a /etc/pacman.conf\n"
            "    sudo pacman-key --add %s\n"
            "    sudo pacman-key --lsign-key %s\n"
            "    sudo pacman -Syu %s\n"
            "To revert these changes:\n"
            "    sudo pacman -R %s\n"
            "    sudo pacman-key --delete %s\n"
            "    sudo mv /etc/pacman.conf.bu /etc/pacman.conf",
            app_name, path('target/repo'),
            path('src/sign/linux/public-key.gpg'), gpg_key, pkg_name, pkg_name,
            gpg_key,
            extra={'wrap': False}
        )
    else:
        assert is_fedora()
        from fbs.repo.fedora import create_repo_fedora
        create_repo_fedora()
        _LOG.info(
            "Done. You can test the repository with the following commands:\n"
            "    sudo rpm -v --import %s\n"
            "    sudo dnf config-manager --add-repo file://%s/target/repo\n"
            "    sudo dnf install %s\n"
            "To revert these changes:\n"
            "    sudo dnf remove %s\n"
            "    sudo rm /etc/yum.repos.d/*%s*.repo\n"
            "    sudo rpm --erase gpg-pubkey-%s",
            path('src/sign/linux/public-key.gpg'), SETTINGS['project_dir'],
            pkg_name, pkg_name, app_name, gpg_key[-8:].lower(),
            extra={'wrap': False}
        )

def _repo_is_supported():
    return is_ubuntu() or is_arch_linux() or is_fedora()

@command
def upload():
    """
    Upload installer and repository to fbs.sh
    """
    require_existing_project()
    try:
        username = SETTINGS['fbs_user']
        password = SETTINGS['fbs_pass']
    except KeyError as e:
        raise FbsError(
            'Could not find setting "%s". You may want to invoke one of the '
            'following:\n'
            ' * fbs register\n'
            ' * fbs login'
            % (e.args[0],)
        ) from None
    _upload_repo(username, password)
    app_name = SETTINGS['app_name']
    url = lambda p: 'https://fbs.sh/%s/%s/%s' % (username, app_name, p)
    message = 'Done! '
    pkg_name = app_name.lower()
    installer_url = url(SETTINGS['installer'])
    if is_linux():
        message += 'Your users can now install your app via the following ' \
                   'commands:\n'
        format_commands = lambda *cmds: '\n'.join('    ' + c for c in cmds)
        repo_url = url(SETTINGS['repo_subdir'])
        if is_ubuntu():
            message += format_commands(
                "sudo apt-get install apt-transport-https",
                "wget -qO - %s | sudo apt-key add -" % url('public-key.gpg'),
                "echo 'deb [arch=amd64] %s stable main' | " % repo_url +
                "sudo tee /etc/apt/sources.list.d/%s.list" % pkg_name,
                "sudo apt-get update",
                "sudo apt-get install " + pkg_name
            )
            message += '\nIf they already have your app installed, they can ' \
                       'force an immediate update via:\n'
            message += format_commands(
                'sudo apt-get update '
                '-o Dir::Etc::sourcelist="/etc/apt/sources.list.d/%s.list" '
                '-o Dir::Etc::sourceparts="-" -o APT::Get::List-Cleanup="0"'
                % pkg_name,
                'sudo apt-get install --only-upgrade ' + pkg_name
            )
        elif is_arch_linux():
            message += format_commands(
                "curl -O %s && " % url('public-key.gpg') +
                "sudo pacman-key --add public-key.gpg && " +
                "sudo pacman-key --lsign-key %s && " % SETTINGS['gpg_key'] +
                "rm public-key.gpg",
                "echo -e '\\n[%s]\\nServer = %s' | sudo tee -a /etc/pacman.conf"
                % (app_name, repo_url),
                "sudo pacman -Syu " + pkg_name
            )
            message += '\nIf they already have your app installed, they can ' \
                       'force an immediate update via:\n'
            message += format_commands('sudo pacman -Syu --needed ' + pkg_name)
        elif is_fedora():
            message += format_commands(
                "sudo rpm -v --import " + url('public-key.gpg'),
                "sudo dnf config-manager --add-repo %s/%s.repo"
                % (repo_url, app_name),
                "sudo dnf install " + pkg_name
            )
            message += "\n(On CentOS, replace 'dnf' by 'yum' and " \
                       "'dnf config-manager' by 'yum-config-manager'.)"
            message += '\nIf they already have your app installed, they can ' \
                       'force an immediate update via:\n'
            message += \
                format_commands('sudo dnf upgrade %s --refresh' % pkg_name)
            message += '\nThis is for Fedora. For CentOS, use:\n'
            message += format_commands(
                'sudo yum clean all && sudo yum upgrade ' + pkg_name
            )
        else:
            raise FbsError('This Linux distribution is not supported.')
        message += '\nFinally, your users can also install without automatic ' \
                   'updates by downloading:\n    ' + installer_url
        extra = {'wrap': False}
    else:
        message += 'Your users can now download and install %s.' % installer_url
        extra = None
    _LOG.info(message, extra=extra)

@command
def release(version=None):
    """
    Bump version and run clean,freeze,...,upload
    """
    require_existing_project()
    if version is None:
        curr_version = SETTINGS['version']
        next_version = _get_next_version(curr_version)
        release_version = prompt_for_value(
            'Release version', default=next_version
        )
    elif version == 'current':
        release_version = SETTINGS['version']
    else:
        release_version = version
    if not is_valid_version(release_version):
        if not is_valid_version(version):
            raise FbsError(
                'The release version of your app is invalid. It should be '
                'three\nnumbers separated by dots, such as "1.2.3". '
                'You have: "%s".' % release_version
            )
    activate_profile('release')
    SETTINGS['version'] = release_version
    log_level = _LOG.level
    if log_level == logging.NOTSET:
        _LOG.setLevel(logging.WARNING)
    try:
        clean()
        freeze()
        if is_windows() and _has_windows_codesigning_certificate():
            sign()
        installer()
        if (is_windows() and _has_windows_codesigning_certificate()) or \
            is_arch_linux() or is_fedora():
            sign_installer()
        if _repo_is_supported():
            repo()
    finally:
        _LOG.setLevel(log_level)
    upload()
    base_json = 'src/build/settings/base.json'
    update_json(path(base_json), { 'version': release_version })
    _LOG.info('Also, %s was updated with the new version.', base_json)

@command
def test():
    """
    Execute your automated tests
    """
    require_existing_project()
    sys.path.append(path('src/main/python'))
    suite = TestSuite()
    test_dirs = SETTINGS['test_dirs']
    for test_dir in map(path, test_dirs):
        sys.path.append(test_dir)
        try:
            dir_names = listdir(test_dir)
        except FileNotFoundError:
            continue
        for dir_name in dir_names:
            dir_path = join(test_dir, dir_name)
            if isfile(join(dir_path, '__init__.py')):
                suite.addTest(defaultTestLoader.discover(
                    dir_name, top_level_dir=test_dir
                ))
    has_tests = bool(list(suite))
    if has_tests:
        TextTestRunner().run(suite)
    else:
        _LOG.warning(
            'No tests found. You can add them to:\n * '+
            '\n * '.join(test_dirs)
        )

@command
def clean():
    """
    Remove previous build outputs
    """
    try:
        rmtree(path('target'))
    except FileNotFoundError:
        return
    except OSError:
        # In a docker container, target/ may be mounted so we can't delete it.
        # Delete its contents instead:
        for f in listdir(path('target')):
            fpath = join(path('target'), f)
            if isdir(fpath):
                rmtree(fpath, ignore_errors=True)
            elif isfile(fpath):
                remove(fpath)
            elif islink(fpath):
                unlink(fpath)

def _has_windows_codesigning_certificate():
    assert is_windows()
    from fbs.sign.windows import _CERTIFICATE_PATH
    return exists(path(_CERTIFICATE_PATH))

def _has_module(name):
    return bool(find_spec(name))

def _get_next_version(version):
    version_parts = version.split('.')
    next_patch = str(int(version_parts[-1]) + 1)
    return '.'.join(version_parts[:-1]) + '.' + next_patch


from os.path import relpath, join

import os

def upload_folder_contents(dir_path, dest_path, bucket, key, secret):
    result = []
    for file_path in _iter_files_recursive(dir_path):
        file_relpath = relpath(file_path, dir_path)
        # Replace backslashes on Windows by forward slashes:
        file_relpath = file_relpath.replace(os.sep, '/')
        file_dest = dest_path + '/' + file_relpath
        upload_file(file_path, file_dest, bucket, key, secret)
        result.append(file_dest)
    return result

def upload_file(file_path, dest_path, bucket, key, secret):
    # Import late to not make boto3 a dependency of fbs only when the upload
    # functionality is actually used.
    import boto3
    s3 = boto3.resource(
        's3', aws_access_key_id=key, aws_secret_access_key=secret
    )
    s3.Bucket(bucket).upload_file(file_path, dest_path)

def _iter_files_recursive(dir_path):
    for subdir_path, dir_names, file_names in os.walk(dir_path):
        for file_name in file_names:
            yield join(subdir_path, file_name)


from fbs import path, _server
from fbs.builtin_commands import prompt_for_value, require_existing_project
from fbs.builtin_commands._util import update_json, SECRET_JSON
from fbs.cmdline import command

import logging

_LOG = logging.getLogger(__name__)

@command
def register():
    """
    Create an account for uploading your files
    """
    require_existing_project()
    username = prompt_for_value('Username')
    email = prompt_for_value('Real email')
    password = prompt_for_value('Password', password=True)
    print('')
    status, text = _server.post_json('register', {
        'username': username, 'email': email, 'password': password
    })
    if status == 201:
        if text:
            _LOG.info(text)
        _login(username, password)
    else:
        _LOG.error('Could not register:\n' + text)

@command
def login():
    """
    Save your account details to secret.json
    """
    require_existing_project()
    username = prompt_for_value('Username')
    password = prompt_for_value('Password', password=True)
    print('')
    _login(username, password)

def _login(username, password):
    update_json(path(SECRET_JSON), {'fbs_user': username, 'fbs_pass': password})
    _LOG.info('Saved your username and password to %s.', SECRET_JSON)


from fbs import path
from fbs.builtin_commands import prompt_for_value, require_existing_project
from fbs.builtin_commands._util import update_json, SECRET_JSON, BASE_JSON
from fbs.cmdline import command

import json
import logging

_LOG = logging.getLogger(__name__)

@command
def init_licensing():
    """
    Generate public/private keys for licensing
    """
    require_existing_project()
    try:
        import rsa
    except ImportError:
        _LOG.error(
            'Please install Python library `rsa`. Eg. via:\n'
            '    pip install rsa'
        )
        return
    nbits = _prompt_for_nbits()
    print('')
    pubkey, privkey = rsa.newkeys(nbits)
    pubkey_args = {'n': pubkey.n, 'e': pubkey.e}
    privkey_args = {
        attr: getattr(privkey, attr) for attr in ('n', 'e', 'd', 'p', 'q')
    }
    update_json(path(SECRET_JSON), {
        'licensing_privkey': privkey_args,
        'licensing_pubkey': pubkey_args
    })
    try:
        with open(path(BASE_JSON)) as f:
            user_base_settings = json.load(f)
    except FileNotFoundError:
        user_base_settings = {}
    public_settings = user_base_settings.get('public_settings', [])
    if 'licensing_pubkey' not in public_settings:
        public_settings.append('licensing_pubkey')
        update_json(path(BASE_JSON), {'public_settings': public_settings})
        updated_base_json = True
    else:
        updated_base_json = False
    message = 'Saved a public/private key pair for licensing to:\n    %s.\n' \
              % SECRET_JSON
    if updated_base_json:
        message += 'Also added "licensing_pubkey" to "public_settings" in' \
                   '\n    %s.\n' \
                   '(This lets your app read the public key when it runs.)\n' \
                   % BASE_JSON
    message += '\nFor details on how to implement licensing for your ' \
               'application, see:\n '\
               '    https://build-system.fman.io/manual#licensing.'
    _LOG.info(message)

def _prompt_for_nbits():
    while True:
        nbits_str = prompt_for_value('Bit size', default='2048')
        try:
            return int(nbits_str)
        except ValueError:
            continue


from fbs import SETTINGS
from fbs_runtime import FbsError
from subprocess import run, DEVNULL, check_call, check_output, PIPE, \
    CalledProcessError

import re

def preset_gpg_passphrase():
    # Ensure gpg-agent is running:
    run(
        ['gpg-agent', '--daemon', '--use-standard-socket', '-q'],
        stdout=DEVNULL, stderr=DEVNULL
    )
    gpg_key = SETTINGS['gpg_key']
    try:
        keygrip = _get_keygrip(gpg_key)
    except GpgDoesNotSupportKeygrip:
        # Old GPG versions don't support keygrips; They use the fingerprint
        # instead:
        keygrip = gpg_key
    check_call([
        SETTINGS['gpg_preset_passphrase'], '--preset', '--passphrase',
        SETTINGS['gpg_pass'], keygrip
    ], stdout=DEVNULL)

def _get_keygrip(pubkey_id):
    try:
        output = check_output(
            ['gpg2', '--with-keygrip', '-K', pubkey_id],
            universal_newlines=True, stderr=PIPE
        )
    except CalledProcessError as e:
        if 'invalid option "--with-keygrip"' in e.stderr:
            raise GpgDoesNotSupportKeygrip() from None
        elif 'No secret key' in e.stderr:
            raise FbsError(
                "GPG could not read your key for code signing. Perhaps you "
                "don't want\nto run this command here, but after:\n"
                "    fbs runvm {ubuntu|fedora|arch}"
            )
        raise
    pure_signing_subkey = _find_keygrip(output, 'S')
    if pure_signing_subkey:
        return pure_signing_subkey
    any_signing_key = _find_keygrip(output, '[^]]*S[^]]*')
    if any_signing_key:
        return any_signing_key
    raise RuntimeError('Keygrip not found. Output was:\n' + output)

def _find_keygrip(gpg2_output, type_re):
    lines = gpg2_output.split('\n')
    for i, line in enumerate(lines):
        if re.match(r'.*\[%s\]$' % type_re, line):
            for keygrip_line in lines[i + 1:]:
                m = re.match(r' +Keygrip = ([A-Z0-9]{40})', keygrip_line)
                if m:
                    return m.group(1)

class GpgDoesNotSupportKeygrip(RuntimeError):
    pass


from fbs import path, SETTINGS
from fbs._gpg import preset_gpg_passphrase
from subprocess import check_call, DEVNULL

def sign_installer_arch():
    installer = path('target/${installer}')
    # Prevent GPG from prompting us for the passphrase when signing:
    preset_gpg_passphrase()
    check_call(
        ['gpg', '--batch', '--yes', '-u', SETTINGS['gpg_key'],
        '--output', installer + '.sig', '--detach-sig', installer],
        stdout=DEVNULL
    )


from fbs.freeze.linux import freeze_linux, remove_shared_libraries

def freeze_ubuntu(debug=False):
    freeze_linux(debug)
    # When we build on Ubuntu on 14.04 and run on 17.10, the app fails to start
    # with the following error:
    #
    #  > This application failed to start because it could not find or load the
    #  > Qt platform plugin "xcb" in "". Available platform plugins are:
    #  > eglfs, linuxfb, minimal, minimalegl, offscreen, vnc, xcb.
    #
    # Interestingly, the error does not occur when building on Ubuntu 16.04.
    # The difference between the two build outputs seems to be
    # libgpg-error.so.0. Removing it fixes the problem:
    remove_shared_libraries('libgpg-error.so.*')
    # libgtk-3.so is present on every Ubuntu system. Make sure we don't ship it
    # to avoid incompatibilities. In particular, running the frozen app with
    # libgtk-3.so from Ubuntu 14 on Ubuntu 16 produces many Gtk warnings
    # "Theme parsing error".
    remove_shared_libraries('libgtk-3.so.*')
    # We also don't want to ship libgio-2.0.so because it is usually present.
    # What's more, if we ship libgio without libgtk, then segmentation faults
    # occur when freezing on Ubuntu 14 and running on Ubuntu 16. The reason for
    # this is that libgio depends on libgtk. Because we don't ship libgtk, this
    # loads the user's libgtk, which is incompatible between Ubuntu 14 and 16.
    remove_shared_libraries('libgio-2.0.so.*')


from fbs import path, SETTINGS
from fbs._state import LOADED_PROFILES
from fbs.resources import _copy
from fbs_runtime._fbs import filter_public_settings
from fbs_runtime._source import default_path
from fbs_runtime.platform import is_mac
from os import rename, makedirs
from os.path import join, dirname
from pathlib import PurePath
from subprocess import run

import fbs_runtime._frozen

def run_pyinstaller(extra_args=None, debug=False):
    if extra_args is None:
        extra_args = []
    app_name = SETTINGS['app_name']
    # Would like log level WARN when not debugging. This works fine for
    # PyInstaller 3.3. However, for 3.4, it gives confusing warnings
    # "hidden import not found". So use ERROR instead.
    log_level = 'DEBUG' if debug else 'ERROR'
    args = [
        'pyinstaller',
        '--name', app_name,
        '--noupx',
        '--log-level', log_level,
        '--noconfirm'
    ]
    for hidden_import in SETTINGS['hidden_imports']:
        args.extend(['--hidden-import', hidden_import])
    args.extend(SETTINGS.get('extra_pyinstaller_args', []))
    args.extend(extra_args)
    args.extend([
        '--distpath', path('target'),
        '--specpath', path('target/PyInstaller'),
        '--workpath', path('target/PyInstaller')
    ])
    args.extend(['--additional-hooks-dir', join(dirname(__file__), 'hooks')])
    if debug:
        args.extend(['--debug', 'all'])
        if is_mac():
            # Force generation of an .app bundle. Otherwise, PyInstaller skips
            # it when --debug is given.
            args.append('-w')
    hook_path = _generate_runtime_hook()
    args.extend(['--runtime-hook', hook_path])
    args.append(path(SETTINGS['main_module']))
    run(args, check=True)
    output_dir = path('target/' + app_name + ('.app' if is_mac() else ''))
    freeze_dir = path('${freeze_dir}')
    # In most cases, rename(src, dst) silently "works" when src == dst. But on
    # some Windows drives, it raises a FileExistsError. So check src != dst:
    if PurePath(output_dir) != PurePath(freeze_dir):
        rename(output_dir, freeze_dir)

def _generate_runtime_hook():
    makedirs(path('target/PyInstaller'), exist_ok=True)
    module = fbs_runtime._frozen
    hook_path = path('target/PyInstaller/fbs_pyinstaller_hook.py')
    with open(hook_path, 'w') as f:
        # Inject public settings such as "version" into the binary, so
        # they're available at run time:
        f.write('\n'.join([
            'import importlib',
            'module = importlib.import_module(%r)' % module.__name__,
            'module.BUILD_SETTINGS = %r' % filter_public_settings(SETTINGS)
        ]))
    return hook_path

def _generate_resources():
    """
    Copy the data files from src/main/resources to ${freeze_dir}.
    Automatically filters files mentioned in the setting files_to_filter:
    Placeholders such as ${app_name} are automatically replaced by the
    corresponding setting in files on that list.
    """
    freeze_dir = path('${freeze_dir}')
    if is_mac():
        resources_dest_dir = join(freeze_dir, 'Contents', 'Resources')
    else:
        resources_dest_dir = freeze_dir
    for path_fn in default_path, path:
        for profile in LOADED_PROFILES:
            _copy(path_fn, 'src/main/resources/' + profile, resources_dest_dir)
            _copy(path_fn, 'src/freeze/' + profile, freeze_dir)


from fbs import path
from fbs._gpg import preset_gpg_passphrase
from subprocess import check_call, DEVNULL

def sign_installer_fedora():
    # Prevent GPG from prompting us for the passphrase when signing:
    preset_gpg_passphrase()
    check_call(
        ['rpm', '--addsign', path('target/${installer}')], stdout=DEVNULL
    )


from urllib.error import HTTPError
from urllib.request import Request, urlopen

import json

_API_URL = 'https://build-system.fman.io/api/'

def post_json(path, data, encoding='utf-8'):
    # We could just use the requests library. But it doesn't pay off to add the
    # whole dependency for just this one function.
    request = Request(_API_URL + path)
    request.add_header('Content-Type', 'application/json; charset=' + encoding)
    data_bytes = json.dumps(data).encode(encoding)
    request.add_header('Content-Length', len(data_bytes))
    try:
        with urlopen(request, data_bytes) as response:
            return response.getcode(), response.read().decode(encoding)
    except HTTPError as e:
        return e.getcode(), e.fp.read().decode(encoding)


from glob import glob
from os.path import dirname, relpath, join

import PySide2

support = None

if PySide2.__version__ == "5.12.2":
    import PySide2.support as support
else:
    try:
        # PySide2 < 5.12.2
        import shiboken2.support as support
    except ImportError:
        # PySide2 >= 5.12.3
        pass

if support is not None:
    """
    This should give roughly the same results as:
    
        from PyInstaller.utils.hooks import collect_data_files
        datas = collect_data_files(
            'shiboken2', include_py_files=True, subdir='support'
        )
    
    The reason we don't do it this way is that it would add a dynamic link to
    PyInstaller, and thus force the GPL on fbs, preventing it from being
    licensed under different terms (such as a commercial license).
    """
    _base_dir = dirname(support.__file__)
    _python_files = glob(join(_base_dir, '**', '*.py'), recursive=True)
    _site_packages = dirname(dirname(_base_dir))
    datas = [(f, relpath(dirname(f), _site_packages)) for f in _python_files]

    hiddenimports = ['typing']


from fbs import path, SETTINGS
from fbs.builtin_commands import require_existing_project
from fbs.cmdline import command
from fbs.resources import _copy
from fbs_runtime import FbsError
from fbs_runtime._source import default_path
from os import listdir
from os.path import exists
from shutil import rmtree
from subprocess import run, CalledProcessError, PIPE

import logging

__all__ = ['buildvm', 'runvm']

_LOG = logging.getLogger(__name__)

@command
def buildvm(name):
    """
    Build a Linux VM. Eg.: buildvm ubuntu
    """
    require_existing_project()
    build_dir = path('target/%s-docker-image' % name)
    if exists(build_dir):
        rmtree(build_dir)
    src_root = 'src/build/docker'
    available_vms = set(listdir(default_path(src_root)))
    if exists(path(src_root)):
        available_vms.update(listdir(path(src_root)))
    if name not in available_vms:
        raise FbsError(
            'Could not find %s. Available VMs are:%s' %
            (name, ''.join(['\n * ' + vm for vm in available_vms]))
        )
    src_dir = src_root + '/' + name
    for path_fn in default_path, path:
        _copy(path_fn, src_dir, build_dir)
    settings = SETTINGS['docker_images'].get(name, {})
    for path_fn in default_path, path:
        for p in settings.get('build_files', []):
            _copy(path_fn, p, build_dir)
    args = ['build', '--pull', '-t', _get_docker_id(name), build_dir]
    for arg, value in settings.get('build_args', {}).items():
        args.extend(['--build-arg', '%s=%s' % (arg, value)])
    try:
        _run_docker(
            args, check=True, stdout=PIPE, stderr=PIPE,
            universal_newlines=True
        )
    except CalledProcessError as e:
        if '/private-key.gpg: no such file or directory' in e.stderr:
            message = 'Could not find private-key.gpg. Maybe you want to ' \
                      'run:\n    fbs gengpgkey'
        else:
            message = e.stdout + '\n' + e.stderr
        raise FbsError(message)
    _LOG.info('Done. You can now execute:\n    fbs runvm ' + name)

@command
def runvm(name):
    """
    Run a Linux VM. Eg.: runvm ubuntu
    """
    args = ['run', '-it']
    for item in _get_docker_mounts(name).items():
        args.extend(['-v', '%s:%s' % item])
    docker_id = _get_docker_id(name)
    args.append(docker_id)
    try:
        _run_docker(args, stderr=PIPE, universal_newlines=True, check=True)
    except CalledProcessError as e:
        if 'Unable to find image' in e.stderr:
            raise FbsError(
                'Docker could not find image %s. You may want to run:\n'
                '    fbs buildvm %s' % (docker_id, name)
            )

def _run_docker(args, **kwargs):
    try:
        return run(['docker'] + args, **kwargs)
    except FileNotFoundError:
        raise FbsError(
            'fbs could not find Docker. Is it installed and on your PATH?'
        )

def _get_docker_id(name):
    prefix = SETTINGS['app_name'].replace(' ', '_').lower()
    suffix = name.lower()
    return prefix + '/' + suffix

def _get_docker_mounts(name):
    result = {'target/' + name.lower(): 'target'}
    # These directories are created inside the container by `buildvm`:
    ignore = {'target', 'venv'}
    for file_name in listdir(path('.')):
        if file_name in ignore:
            continue
        result[file_name] = file_name
    path_in_docker = lambda p: '/root/%s/%s' % (SETTINGS['app_name'], p)
    return {path(src): path_in_docker(dest) for src, dest in result.items()}

def _get_settings(name):
    return SETTINGS['docker_images'][name]


# UTF-8
#
# For more details about fixed file info 'ffi' see:
# http://msdn.microsoft.com/en-us/library/ms646997.aspx

VSVersionInfo(
  ffi=FixedFileInfo(
    # filevers and prodvers should be always a tuple with four items: (1, 2, 3, 4)
    # Set not needed items to zero 0. Must always contain 4 elements.
    filevers=(
        int('${version}'.split('.')[0]),
        int('${version}'.split('.')[1]),
        int('${version}'.split('.')[2]),
        0
    ),
    prodvers=(
        int('${version}'.split('.')[0]),
        int('${version}'.split('.')[1]),
        int('${version}'.split('.')[2]),
        0
    ),
    # Contains a bitmask that specifies the valid bits 'flags'r
    mask=0x3f,
    # Contains a bitmask that specifies the Boolean attributes of the file.
    flags=0x0,
    # The operating system for which this file was designed.
    # 0x4 - NT and there is no need to change it.
    OS=0x40004,
    # The general type of file.
    # 0x1 - the file is an application.
    fileType=0x1,
    # The function of the file.
    # 0x0 - the function is not defined for this fileType
    subtype=0x0,
    # Creation date and time stamp.
    date=(0, 0)
    ),
  kids=[
    StringFileInfo(
      [
      StringTable(
        '040904B0',
        [StringStruct('CompanyName', '${author}'),
        StringStruct('FileDescription', '${app_name}'),
        StringStruct('FileVersion', '${version}.0'),
        StringStruct('InternalName', '${app_name}'),
        StringStruct('LegalCopyright', '© ${author}. All rights reserved.'),
        StringStruct('OriginalFilename', '${app_name}.exe'),
        StringStruct('ProductName', '${app_name}'),
        StringStruct('ProductVersion', '${version}.0')])
      ]),
    VarFileInfo([VarStruct('Translation', [1033, 1200])])
  ]
)


from fbs import path
from fbs.resources import copy_with_filtering
from fbs_runtime._source import default_path
from os import makedirs, rename
from os.path import exists
from shutil import rmtree, copy
from subprocess import check_call, DEVNULL

def create_repo_fedora():
    if exists(path('target/repo')):
        rmtree(path('target/repo'))
    makedirs(path('target/repo/${version}'))
    copy(path('target/${installer}'), path('target/repo/${version}'))
    check_call(['createrepo_c', '.'], cwd=(path('target/repo')), stdout=DEVNULL)
    repo_file = path('src/repo/fedora/${app_name}.repo')
    use_default = not exists(repo_file)
    if use_default:
        repo_file = default_path('src/repo/fedora/AppName.repo')
    copy_with_filtering(
        repo_file, path('target/repo'), files_to_filter=[repo_file]
    )
    if use_default:
        rename(
            path('target/repo/AppName.repo'),
            path('target/repo/${app_name}.repo')
        )


from fbs_runtime.excepthook import ExceptionHandler
from fbs_runtime.excepthook._util import RateLimiter

class SentryExceptionHandler(ExceptionHandler):
    def __init__(
        self, dsn, app_version, environment, callback=lambda: None,
        rate_limit=10
    ):
        raise RuntimeError(
            'Error tracking via Sentry is only available in fbs Pro. '
            'Please obtain it from https://build-system.fman.io/pro.'
        )